
:- [util].
:- [reason].
:- [racer].
:- [reado].
:- [realis].
:- [haslex].
:- [process_axioms].
:- [known].
:- [refine].
:- [owl].

% Execution flags

:- dynamic(att_tracing/0).

negations.      % whether negations allowed
optimisations.  % whether optimisations using subsumption are used
att_tracing.  % whether to trace operations during attributive finding

go :- 
   reado,                                                     % read in an ontology 
   new_kb(system),                                            % set up...
   addaxioms(system),                                         %  ... system KB
   new_kb(user),                                              % set up...
   addvocab(user),                                            %  ... user KB
   write('What is a/an ...? '), ttyflush,                     % get target concept
   read(Tar),
   write('Limit: '), ttyflush, read(Lim),                     % get complexity limit
   write('Viewpoint (user/system): '), ttyflush, read(EGO),   % get viewpoint
   attributives(Tar,k(system,user,EGO),Lim,Subs),             % get subsumers
   nl,
   forall(member(X,Subs),tell_user(subclass_of(Tar,X))),      % realise them as language
   nl, !.

:- dynamic filtered_subsumer/2.    

% attributives(+Concept,+KBs,+Complexity,-List).
%
% Concept is the name of the concept whose subsumers are sought
% KBs is a structure k(Sys,User,View) where Sys is the name of the
%   system KB, User the name of the user KB and View is the name of the viewpoint KB
%   (called the "ego" in some of the code)
%   The names are passed down to the Racer interface code. They are associated
%   with ports etc in dig/dig_client_setting.pl (predicate dig_client_init)
% Complexity is the maximum syntactic complexity allowed in subsumers
% List returns the list of subsumers
%
% df is the predicate that does the real work

attributives(X,KBs,Compl,List) :-
   (att_tracing -> (write('Seeking attributes for '), write(X), nl); true),
   rec(target,X),
   rec(limit,Compl),
   retractall(racercount(_)),
   assert(racercount(0)),
   find_relevant(X,[],Rel),
   system_KB(KBs,Sys),
   (att_tracing -> time(df(subs(Sys,X),KBs,Compl,Rel,Cands)); df(subs(Sys,X),KBs,Compl,Rel,Cands)),
   length(Cands,N),  racercount(CT),
   (att_tracing -> (nl, write(N), write(' subsumers for '), write(X), write(' found'), nl,
        write(CT), write(' calls to Racer'), nl, nl); 
        (write('I can see '), write(N), write(' possibilities. Let\'s see...'),nl)),
   rec(cands,N), rec(calls,CT),
   % doall((member(CC,Cands),lclass(CC),nl)), 
   dofilter(X,KBs,Cands), 
   findall(ZZ,retract(filtered_subsumer(X,ZZ)),List).
   % length(List,FL),
   % rec(filtered,FL).

:- dynamic(reduction/3).

% 1. Remove things known to user 
dofilter(Tar,KBs,Cands) :-
   retractall(filtered_subsumer(Tar,_)),
   user_KB(KBs,UKB),
   findall(X,(member(X,Cands),\+ subsumes(UKB,X,Tar)),L),
   doall((member(X1,L),assert(filtered_subsumer(Tar,X1)))),
   fail.
% 2a. Reduce using local ranges
dofilter(Tar,_,_) :-
   (att_tracing -> (write('reducing using local ranges'),nl); true),
   filtered_subsumer(Tar,X), filtered_subsumer(Tar,Y), \+ X=Y,
   reduces(X,Y,YNew),
   retract(filtered_subsumer(Tar,Y)),
   assert(filtered_subsumer(Tar,YNew)),
   fail.
% 2b. Reduce redundant negation/conjunctions
dofilter(Tar,_,_) :-
   (att_tracing -> (write('reducing conjunctions with negations'),nl); true),
   conjneg(Tar), fail.
% 3. Second stage of pruning subsuming candidates
dofilter(Tar,KBs,_) :-
%   listing(filtered_subsumer),
%   write('filter 2'), nl,
   (att_tracing -> (write('filtering...'),nl); true),
   ego_KB(KBs,KB),
   user_KB(KBs,UKB),
   setof(X,(retract(filtered_subsumer(Tar,X)),\+ subsumes(UKB,X,Tar)),L),   % Remove duplicates from reductions
   (att_tracing -> time(most_specific(L,KB,L1)); most_specific(L,KB,L1)), 
   doall((member(X1,L1),assert(filtered_subsumer(Tar,X1)))),
   (att_tracing -> (length(L1,N), write(N), write(' filtered subsumers found'), nl,
      doall((filtered_subsumer(Tar,X),(wax(X),nl)))); true), 
   fail.
dofilter(X,KBs,_) :-
%   nl, write('checking...'), nl,
   system_KB(KBs,SKB),
   filtered_subsumer(X,Y),
   \+ subsumes(SKB,Y,X),
   write('***ERROR IN SUBSUMER '), write(Y), nl,
   fail.
dofilter(_,_,_).
%   write(ok), nl.

% Remove conjuncts which subsume others

% 1. Version using classification
most_specific(L,KB,L1) :-
   findall((X,NX),(member(X,L),tostring(X,NX),addalias(KB,X,NX)),Lpair),
   gensym(Conj),
   addalias(KB,intersection_of(L),Conj),
   all_parents(KB,Conj,Pars),
   findall(X,(member(NX,Pars),once(member((X,NX),Lpair))),L1).

/*
% 2. Version comparing pairwise
most_specific(L,KB,L1) :-
   fil1(L,[],KB,L1).
*/

fil1([X|After],Sofar,KB,L1) :- member(Y,Sofar), subsumes(KB,X,Y), !,
   %wax(X), write(' subsumes '), wax(Y), nl, 
   fil1(After,Sofar,KB,L1).
fil1([X|After],Sofar,KB,L1) :- member(Y,After), subsumes(KB,X,Y),
   ((\+ subsumes(KB,Y,X)); (size(X,XS),size(Y,YS),YS=<XS)), !,
   %wax(X), write(' subsumes '), wax(Y), nl,  
   fil1(After,Sofar,KB,L1).
fil1([X|L],Sofar,KB,L1) :- !, fil1(L,[X|Sofar],KB,L1).
fil1([],L,_,L).

/*
% 3. Better version, but relies on subsumption test wrt remaining intersection:
most_specific(L,KB,L2) :- select(X,L,L1), subsumes(KB,X,intersection_of(L1)), !,
   % wax(X), write(' subsumes '), wax(intersection_of(L1)), nl,
   most_specific(L1,KB,L2).
most_specific(L,_,L).
*/

% Remove subsumees from a list

most_general(L,KB,L1) :-
   gen1(L,[],KB,L1).

gen1([X|After],Sofar,KB,L1) :- member(Y,Sofar), subsumes(KB,Y,X), !,
   % wax(Y), write(' subsumes '), wax(X), nl, 
   gen1(After,Sofar,KB,L1).
gen1([X|After],Sofar,KB,L1) :- member(Y,After), subsumes(KB,Y,X),
   ((\+ subsumes(KB,X,Y)); (size(X,XS),size(Y,YS),YS=<XS)), !,
   % wax(Y), write(' subsumes '), wax(X), nl,  
   gen1(After,Sofar,KB,L1).
gen1([X|L],Sofar,KB,L1) :- !, gen1(L,[X|Sofar],KB,L1).
gen1([],L,_,L).

% reduction wrt local ranges

reduces(restriction(P,L),Y,YNew) :-
   member(all_values_from(C),L),
   red1(P,C,Y,YNew), \+ Y=YNew,
   (att_tracing -> (wax(Y), write(' reduces to '), wax(YNew), write(' via '), 
		    wax(restriction(P,[all_values_from(C)])), nl); true).

red1(P,intersection_of([C|L]),Y,YNew) :- !,
   red1(P,C,Y,Y1),
   red1(P,intersection_of(L),Y1,YNew).
red1(P,C,restriction(P,L),restriction(P,[some_values_from(CRest)|LRest])) :-
   select(some_values_from(C1),L,LRest),
   aconj(C,C1,CRest), !.
red1(P,C,restriction(P,L),restriction(P,[some_values_from(CNew)|L1])) :-
   select(some_values_from(C1),L,L1),
   reduces(C,C1,CNew).
red1(_,_,Y,Y).

   aconj(C,C,'TOP').
   aconj(C,intersection_of(L),Rest) :-
      select(C,L,L1), newconj(L1,Rest).

   newconj([],'TOP') :- !.
   newconj([X],X) :- !.
   newconj(L,intersection_of(L)).

% reduction of conjunctions with negations

conjneg(Tar) :-
   filtered_subsumer(Tar,intersection_of([Neg,Pos1])),
   filtered_subsumer(Tar,intersection_of([Neg,Pos2])),
   \+ Pos1=Pos2,
   retract(filtered_subsumer(Tar,intersection_of([Neg,Pos2]))), !,
   assert(filtered_subsumer(Tar,Pos2)),
   (att_tracing -> (wax(intersection_of([Neg,Pos2])), write(' reduces to '), wax(Pos2), nl); true),
   conjneg(Tar).
conjneg(_).

% make something that specifies a set of tests from a name

make_test(Name,X,X-Name).

% Test names:
%
% subs(KB,Tar) - find most specific subsumers of Tar
% subscomplex(KB,Tar) - ditto, but don't allow solutions equivalent to Tar
% notsubs(KB,Tar) - find most general non-subsumers of Tar
% under(KB,Tar) - find most general subsumees of Tar
% gsubs(KB,Tar) - find most general things whose negations subsume Tar

% The goal is to find maximal or minimal x such that p(x)
% p(x) = on_path(x) ^ return_solution(x)
%
% If the goal is to produce maximal x then
%    p(x) ^ y <= x implies p(y)
% If the goal is to produce minimal x then
%    p(x) ^ x <= y implies p(y)

% test which must be true of everything on the path from Top to a solution,
% including the solution. it can also be true of things not on the path.
% on_path('TOP'-X) must be true for all X

on_path(X-subs(KB,Tar)) :- !, \+ X=Tar, subsumes(KB,X,Tar).
on_path(C-subscomplex(KB,Y)) :- !, \+ subsumes(KB,Y,C), subsumes(KB,C,Y).
on_path(_-notsubs(_,_)) :- !.
on_path(X-under(KB,Tar)) :- !, satisfiable(KB,intersection_of([X,Tar])).
on_path(X-gsubs(KB,_)) :- !, \+ subsumes(KB,'BOTTOM',X).

% which tests are finding most GENERAL X?

most_general(_-gsubs(_,_)) :- !.
most_general(_-notsubs(_,_)) :- !.
most_general(_-under(_,_)) :- !.

% test for whether to try looking at descendents for solutions (in which case
% those solutions will win over the current concept), assuming that
% on_path succeeds.

try_descendents(_-subs(_,_)) :- !.
try_descendents(_-subscomplex(_,_)) :- !.
try_descendents(X-notsubs(KB,Tar)) :- !, subsumes(KB,X,Tar).
try_descendents(X-under(KB,Tar)) :- !, (X=Tar; \+ subsumes(KB,Tar,X)), !.
try_descendents(X-gsubs(KB,Tar)) :- !, \+ subsumes(KB,complement_of(X),Tar).

% test for whether the current concept is a solution, given that on_path
% succeeds and (try_descendents fails or there are no solutions in descendents)

return_solution(_-subs(_,_)) :- !.
return_solution(_-subscomplex(_,_)) :- !.
return_solution(X-notsubs(KB,Tar)) :- !, \+ subsumes(KB,X,Tar).
return_solution(X-under(KB,Tar)) :- !, \+ X=Tar, subsumes(KB,Tar,X).
return_solution(X-gsubs(KB,Tar)) :- !, subsumes(KB,complement_of(X),Tar).

% df(+TestName,+KBs,+CompLimit,+Rel,-Cands)
%
% TestName is the name of the test to be applied to possible candidates
%    (converted to an actual test by make_test)
% KBs is the knowledge bases to be used k(System,User,Viewpoint)
% CompLimit is the complexity limit
% Rel is the concept names allowed to be used in the result (e.g. from relevance test)
% Cands is the list of candidates returned

:- dynamic target/1.

df(TestName,KBs,N,Rel,Cands) :-
   (att_tracing -> true; (write(hmm),ttyflush)),
   make_test(TestName,Y,Test),
   findall(Cand,(
      alltop(Y,0,N,NA,[],Test,KBs,Rel),
      (att_tracing -> (write('> '), wax(Y), nl); (write(.),ttyflush)),
      ddf(NA,TestName,KBs,Y,0,Rel,Solns),
      member(Cand,Solns)), Cands),
   (att_tracing -> true; nl).

%ddf(Avail,TestNames,KBs,Curr,Indent,Rel,Solns)
%

ddf(A,TestName,KBs,X,Ind,Rel,Set) :-
   make_test(TestName,X,Test),
   try_descendents(Test), !, 
   Ind1 is Ind+1,
   % tab(Ind), write('> '), wax(X), nl,
   findall(XX,(childd(A,TestName,KBs,X,Ind1,Rel,S1),member(XX,S1)),Local),
   dddf(Local,KBs,TestName,X,Ind,Set).
ddf(_,TestName,KBs,X,Ind,_,[X]) :- make_test(TestName,X,Test), return_solution(Test), 
   \+ redundant(KBs,X), 
   (att_tracing -> (tab(Ind), write('= '), wax(X), nl); true), !.
ddf(_,_,_,_,_,_,[]).
   
childd(A,TestName,KBs,X,Ind,Rel,Set) :-
   step_refine(X,New,A,A1,negonly,Ind,Ind1,TestName,KBs,Rel),
   ddf(A1,TestName,KBs,New,Ind1,Rel,Set).

dddf([],KBs,TestName,X,Ind,[X]) :- make_test(TestName,X,Test), return_solution(Test), 
    \+ redundant(KBs,X), 
   (att_tracing -> (tab(Ind), write('= '), wax(X), nl); true), !.
dddf([],_,_,_,_,[]) :- !.
dddf(L,_,_,_,_,L).
 
%tell the user something

tell_user(although_never(Prev,Tar,X)) :- !,
   best_realise(although_never(subclass_of(Prev,X),subclass_of(Tar,X))),
   assert(user_axiom(subclass_of(Tar,complement_of(X)))).
tell_user(although_not(Prev,Tar,X)) :- !,
   best_realise(although_not(subclass_of(Prev,X),subclass_of(Tar,X))).
   % add individual axiom
tell_user(Ax) :- best_realise(Ax), assert(user_axiom(Ax)).
