% Tracking which concepts are known and what knowledge the user has

% which concepts and roles can be mentioned?
% NB all superconcepts of includable concepts must be includable

% Not sure what this was intended for. down_class and up_class assume that
% if an optimal formula contains a non-includable class then a more
% specific/general one will not be returned. I.e. the set of formulae
% returned is the subset of the optimal ones which only have includable
% classes

includable(_).

% which concepts are not allowed to be described?

known('TOP').
known('BOTTOM').

% Add initial user axioms

initial_user_axioms :-
   all_concepts(Cs),
   all_roles(Rs),
   obvious_top(Cs),
   obvious_ranges(Cs,Rs),
   obvious_subsumptions(Cs).
   % listing(user_axiom).

   obvious_top(Cs) :-
      member(X,['Thing',thing,'THING','Entity']), 
      member(X,Cs),
      equiv(system,'TOP',X),
      assert(user_axiom(equivalent_classes(['TOP',X]))),
      fail.
   obvious_top(_).

   obvious_ranges(Cs,Rs) :-
      member(R,Rs),
      name(R,Na), append("has",Na1,Na), name(C,Na1),
      member(C,Cs),
      equiv(system,'TOP',restriction(R,[all_values_from(C)])),
      assert(user_axiom(equivalent_classes(['TOP',restriction(R,[all_values_from(C)])]))),
      fail.
   obvious_ranges(_,_).

   obvious_subsumptions(Cs) :-
      member(C,Cs),
      name(C,Na),
      member(C1,Cs), \+ C1=C,
      name(C1,Na1), append(_,Na1,Na),
      subsumes(system,C1,C),
      assert(user_axiom(subclass_of(C,C1))), fail.
   obvious_subsumptions(_).
   
