% RACER interface

:- ['dig/dig_client'].

% set this true if you want to add axioms one by one

:- dynamic single_axioms/0.

single_axioms :- fail.    % set this true for debugging only

% counter of Racer calls

:- dynamic racercount/1.
racercount(0).

racer_tell(X,[A|As]) :- single_axioms, !, 
    racer_tell1(X,[A]),
    racer_tell(X,As).
racer_tell(X,[A|As]) :- !, 
   racer_tell1(X,[A|As]).
racer_tell(_,[]) :- !.
racer_tell(X,Ass) :- racer_tell(X,[Ass]).

racer_tell1(Mod,List) :-
   % write('[ '), write(List), nl,
   findall(X,(member(Y,List),xo(Y,X)),XML), 
   dig_tell1(Mod,List,elements(XML),Ans,[]), 
   % write('] '), write(Ans), nl,
   dig_feedback(Ans,List), 
   retract(racercount(N)), N1 is N+1, assert(racercount(N1)), 
   (N=50000 -> (write('****RACER: more than 50000 calls'),nl,break); true), !.

dig_tell1(Mod,_,XML,Ans,Op) :- dig_tell(Mod,XML,Ans,Op), !.
dig_tell1(_,Ass,XML,_,_) :-
   write('DIG tell failed on '), write(XML), write(' for '), write(Ass), nl, fail.

:- dynamic bad/1.

dig_feedback(answer(_,[element(response,_,Elts)]),Ass) :- 
   df1(Elts,Ass).
dig_feedback(Ans,Ass) :-
   write('DIG unexpected response to '), write(Ass), write(' - '), write(Ans), 
   assert(bad(Ans)), nl.

   df1([],_) :- !.
   df1([element(ok,_,[X|Y])|Rest],Ass) :- !,
      write('DIG warning for '), write(Ass), write(' - '), write([X|Y]), nl,
      df1(Rest,Ass).
   df1([element(ok,_,[])|Rest],Ass) :- !,
      df1(Rest,Ass).
   df1([element(error,A,B)|Rest],Ass) :- !,
      write('DIG error for '), write(Ass), write(' - '), write(A), write(' '), write(B), nl,
      df1(Rest,Ass).
   df1([A|Rest],Ass) :- atom(A), df1(Rest,Ass).

% memoing version of racer_ask

racer_ask(X,Query) :- racer_ask(X,Query,R), \+ R=false.

racer_ask(X,Query,Resp) :- memogoal(X,Query,R), !, Resp=R.
racer_ask(X,Query,Resp) :- racer_ask1(X,Query,Resp1), after_ask(X,Query,Resp1,Resp).

   after_ask(X,Query,error,Resp) :- !, racer_ask(X,Query,Resp).   % try again on error
   after_ask(X,Query,Resp1,Resp) :- memopred(X,Query,Resp1,Ass), !, assert(Ass), 
      memogoal(X,Query,Resp), !.

% reset - used when the KB has changed

memoreset(KB) :- retractall(memosubsumes(_,_,KB,_)), retractall(memoquery(_,KB,_)).

% partial reset - used when the KB has changed only by the addition of axioms
% which define new classes - existing subsumption results stand

memopartialreset :- retractall(memoquery(_,_,_)).

:- dynamic memosubsumes/4, memoquery/3.
memopred(M,subsumes(X,Y),R,memosubsumes(X,Y,M,R)) :- !.
memopred(M,all_parents(X),R,memoquery(all_parents(X),M,R1)) :- !, flatten_syns(R,R1).
memopred(M,all_children(X),R,memoquery(all_children(X),M,R1)) :- !, flatten_syns(R,R1).
memopred(M,Q,R,memoquery(Q,M,R)).

memogoal(M,subsumes(X,Y),R) :- !, memosubsumes(X,Y,M,R).
memogoal(M,Q,R) :- memoquery(Q,M,R).

racer_ask1(Mod,Query,Resp) :- 
   % write('[ '), write(Query), nl,
   xq(Query,XML), dig_ask1(Mod,Query,elements([XML]),Ans,[]), 
   dig_response(Ans,Query,Resp), 
   % write('- '), write(Resp), nl,
   retract(racercount(N)), N1 is N+1, assert(racercount(N1)), !,
   (N=50000 -> (write('****RACER: more than 50000 calls'),nl,break); true), !.

dig_ask1(Mod,_,XML,Ans,Op) :- dig_ask(Mod,XML,Ans,Op), !.
dig_ask1(_,Ass,XML,_,_) :-
   write('DIG ask failed on '), write(XML), write(' for '), write(Ass), nl, fail.

% Give a command (DIG 1.1 only)

racer_command(Mod,Cmd,Resp) :- 
   % write('[ '), write(Cmd), nl,
   xq(Cmd,XML), dig_cmd1(Mod,Cmd,XML,Ans,[]), 
   dig_response(Ans,Cmd,Resp), 
   % write('- '), write(Resp), nl,
   retract(racercount(N)), N1 is N+1, assert(racercount(N1)), !.

dig_cmd1(Mod,_,XML,Ans,Op) :- dig_command(Mod,XML,Ans,Op), !.
dig_cmd1(_,Ass,XML,_,_) :-
   write('DIG command failed on '), write(XML), write(' for '), write(Ass), nl, fail.

% Process DIG responses

dig_response(answer(_,[element(responses,_,X)]),Ass,Resp) :- !, digr1(X,Ass,Resp).
dig_response(answer(_,[element(response,_,X)]),Ass,Resp) :- digr1(X,Ass,Resp).

digr1([element(true,_,[])|_],_,true) :- !.
digr1([element(false,_,[])|_],_,false) :- !.
digr1([element(error,A,B)|_],Ass,error) :- !,
   write('DIG error for '), write(Ass), write(' - '), write(A), write(' '), write(B), nl, !, break.
digr1([element(conceptSet,_,List)|_],_,CList) :- clist(List,CList), !.
digr1([element(roleSet,_,List)|_],_,CList) :- clist(List,CList), !.
digr1([element(individualSet,_,List)|_],_,CList) :- clist(List,CList), !.
digr1([element(kb,[uri=N],[])|_],_,uri(N)) :- !.
digr1([element(ok,_,_)|_],_,ok) :- !.
digr1([_,Y|L],Ass,Resp) :- digr1([Y|L],Ass,Resp), !.
digr1(Ans,Ass,_) :- 
   write('DIG unexpected response to '), write(Ass), write(' - '), write(Ans), nl, fail.   

   clist([],[]) :- !.
   clist([element(synonyms,_,L1)|L2],[synonyms(Res)|End]) :- !,
      clist1(L1,Res,[]), clist(L2,End).
   clist([_|L2],Res) :- !,
      clist(L2,Res).

   clist1([],L,L) :- !.
   clist1([element(catom,[name=N],[])|X],[N|L],E) :- !, clist1(X,L,E).
   clist1([element(ratom,[name=N],[])|X],[N|L],E) :- !, clist1(X,L,E).
   clist1([_|X],L,E) :- clist1(X,L,E).

% flatten a list including synonyms (not always needed) - take all from each syn

flatten_syns([synonyms([])|Rest],F) :- !, flatten_syns(Rest,F).
flatten_syns([synonyms([X|L])|Rest],[X|F]) :- !, flatten_syns([synonyms(L)|Rest],F).
flatten_syns([X|Rest],[X|F]) :- !, flatten_syns(Rest,F).
flatten_syns([],[]).

% Convert to XML representation. Returns a set of tells on backtracking.

xo('TOP',element(top,[],[])) :- !.
xo('BOTTOM',element(bottom,[],[])) :- !.
xo(clearKB,element(clearKB,[],[])) :- !.
xo(declare_class(C),element(defconcept,[name=C],[])) :- !.
xo(declare_role(R),element(defrole,[name=R],[])) :- !.
xo(declare_individual(R),element(defindividual,[name=R],[])) :- !.
xo(subclass_of(C1,C2),element(impliesc,[],[C1x,C2x])) :- !,
   xo(C1,C1x), xo(C2,C2x).
xo(equivalent_classes([C1,C2]),element(equalc,[],[C1x,C2x])) :- !,
   xo(C1,C1x), xo(C2,C2x).
xo(disjoint_classes(Cs),element(disjoint,[],Csx)) :- !,
   xol(Cs,Csx).
xo(union_of(Cs),element(or,[],Csx)) :- !,
   xol(Cs,Csx).
xo(intersection_of(Cs),element(and,[],Csx)) :- !,
   xol(Cs,Csx).
xo(complement_of(C),element(not,[],[Cx])) :- !,
   xo(C,Cx).
xo(one_of(List),element(iset,[],L)) :-
   xoil(List,L), !.
xo(restriction(R,L),element(and,[],Lx)) :- 
   xoresl(R,L,Lx), !.
xo(object_property(_,[]),_) :- !, fail.
xo(object_property(R,[X|L]),Px) :- !, member(P,[X|L]), xop(R,P,Px).
xo(C,element(catom,[name=C],[])) :- atomic(C), !.
xo(subproperty_of(P1,P2),element(impliesr,[],[element(ratom,[name=P1],[]),element(ratom,[name=P2],[])])) :- !.
xo(equivalent_properties(List),element(equalr,[],XL)) :- !, xorl(List,XL).
xo(instance_of(I,C),element(instanceof,[],[element(individual,[name=I],[]),element(catom,[name=C],[])])) :- !.
xo(related(I,P,I1),element(related,[],
   [element(individual,[name=I],[]),element(ratom,[name=P],[]),
   element(individual,[name=I1],[])])) :- !.
xo(all_different(L),element(disjoint,[],List)) :- 
   findall(element(iset,[],[element(individual,[name=N],[])]),member(N,L),List), !.
xo(Str,_) :- write('*** CANNOT PUT IN DIG: '), write(Str), nl, fail.

xor(R,all_values_from(D),element(all,[],[element(ratom,[name=R],[]),Dx])) :- !, xo(D,Dx).
xor(R,some_values_from(D),element(some,[],[element(ratom,[name=R],[]),Dx])) :- !, xo(D,Dx).
xor(R,value(I),element(and,[],[
     element(all,[],[element(ratom,[name=R],[]),element(iset,[],[element(individual,[name=I],[])])]),
     element(some,[],[element(ratom,[name=R],[]),element(iset,[],[element(individual,[name=I],[])])])])) :- !.
xor(R,min_cardinality(I),element(atleast,[num=I],[element(ratom,[name=R],[]),element(top,[],[])])) :- !.
xor(R,max_cardinality(I),element(atmost,[num=I],[element(ratom,[name=R],[]),element(top,[],[])])) :- !.
xor(R,cardinality(I),element(and,[],[
   element(atleast,[num=I],[element(ratom,[name=R],[]),element(top,[],[])]),
   element(atmost,[num=I],[element(ratom,[name=R],[]),element(top,[],[])])])) :- !.

xop(R,super(R1),element(impliesr,[],[element(ratom,[name=R],[]),element(ratom,[name=R1],[])])) :- !.
xop(R,inverse_of(R1),element(equalr,[],[element(ratom,[name=R],[]),
   element(inverse,[],[element(ratom,[name=R1],[])])])) :- !.
xop(R,functional,element(functional,[],[element(ratom,[name=R],[])])) :- !.
xop(R,inverse_functional,
   element(equalr,[],[element(inverse,[],[element(ratom,[name=R],[])]),element(ratom,[name=Inv],[])])) :-
   name(R,Na),append("inverse",Na,Na1), name(Inv,Na1).
xop(R,inverse_functional,
   element(functional,[],[element(ratom,[name=Inv],[])])) :- !,
   name(R,Na),append("inverse",Na,Na1), name(Inv,Na1).
xop(R,transitive,element(transitive,[],[element(ratom,[name=R],[])])) :- !.
xop(R,domain(D),element(domain,[],[element(ratom,[name=R],[]),Dx])) :- !, xo(D,Dx).
xop(R,range(D),element(range,[],[element(ratom,[name=R],[]),Dx])) :- !, xo(D,Dx).
xop(R,X,_) :-
   write('*** CANNOT PUT PROPERTY IN DIG: '), write(R-X), nl, fail.

   xol([],[]).
   xol([C|Cs],[Cx|Cxs]) :- xo(C,Cx), xol(Cs,Cxs).

   xoil([],[]).
   xoil([I|Is],[element(individual,[name=I],[])|Ix]) :- xoil(Is,Ix).

   xoresl(_,[],[]).
   xoresl(R,[D|Ds],[Dx|Dxs]) :- xor(R,D,Dx), xoresl(R,Ds,Dxs).

   xorl([R|Rs],[element(ratom,[name=R],[])|X]) :- !, xorl(Rs,X).
   xorl([],[]).

xq(parents(C),element(parents,ID,[X])) :- xo(C,X), !, add_id(ID).
xq(children(C),element(children,ID,[X])) :- xo(C,X), !, add_id(ID).
xq(all_parents(C),element(parents,ID,[X])) :- xo(C,X), !, add_id(ID).
xq(all_children(C),element(children,ID,[X])) :- xo(C,X), !, add_id(ID).
xq(subsumes(X,Y),element(subsumes,ID,[X1,Y1])) :- xo(X,X1), xo(Y,Y1), !, add_id(ID).
xq(equivalents(X),element(equivalents,ID,[XX])) :- xo(X,XX), !, add_id(ID).
xq(all_concepts,element(allConceptNames,ID,[])) :- !, add_id(ID).
xq(all_roles,element(allRoleNames,ID,[])) :- !, add_id(ID).
xq(satisfiable(C),element(satisfiable,ID,[C1])) :- !, xo(C,C1), add_id(ID).
xq(instances(C),element(instances,ID,[C1])) :- !, xo(C,C1), add_id(ID).
xq(newKB,element(newKB,[],[])) :- !.
xq(releaseKB(U),element(releaseKB,[uri=U],[])) :- !.

xq(Q,_) :- write('*** CANNOT PUT IN DIG '), write(Q), nl, fail.

add_id([id=I]) :- dig1p1, !, gensym(id,I).
add_id([]).





