:- use_module(library('semweb/rdf_db')).
:- [util].

:- dynamic(axiom/1).
:- dynamic(transitive/1).

installo(KB) :-
   reado,
   new_kb(KB),
   addaxioms(KB),
   all_parents(KB,'BOTTOM',L),
   write('Pre-bottoms: '), write(L), nl.

reado :-
   whicho(O),
   reado(O).

reado(O) :-
   % rec(ontology,O),
   retractall(axiom(_)),
   retractall(transitive(_)),
   retractall(filtered_subsumer(_,_)),
   retractall(all_roles(_)),
   retractall(all_concepts(_)),   
   rdf_reset_db,
   rdf_load(O),
   rdf_transaction(deal1),
   rdf_transaction(deal2),
   rdf_transaction(deal3),
   (setof(R,axiom(related(_,R,_)),Rels) -> 
    (write('**Individual relations used: '), write(Rels), nl); true),
   ((rdf(XX,_,_),non_anon(XX)) -> (nl, write('*** Unused triples: '), nl, nl,
      doall((
         rdf(X,Y,Z), non_anon(X),
         rdf_global_id(X1,X),
         rdf_global_id(Y1,Y),
         rdf_global_id(Z1,Z),
         write(X1), write('  '), write(Y1), write('  '), write(Z1),nl)),nl);
    true),
   findall(a,axiom(_),L), length(L,N), 
   write('** '), write(N), write(' axioms loaded'), nl, 
   % rec(axioms,N),
   number_axioms,
   once((setof(C,Ax^(axiom(Ax),mowl_axiom(Ax,C,class)),L1); L1=[])),
   length(L1,N1), write('** '), write(N1), write(' classes: '), write(L1), nl,
   assert(all_concepts(L1)), 
   % rec(concepts,N1),
   once((setof(C,Ax^(axiom(Ax),mowl_axiom(Ax,C,role)),L2); L2=[])),
   length(L2,N2), write('** '), write(N2), write(' properties; '), write(L2), nl,
   % rec(properties,N2),
   assert(all_roles(L2)), !.

   % pick an ontology
   whicho(X) :- 
      findall(X,ontol(X),Onts),
      menu_select(Onts,Ont),
      name(Ont,C1), ontol_directory(DirN), append(DirN,C1,C2), name(X,C2).

read_test :-
   ontol(Ont),
   name(Ont,C1), ontol_directory(DirN), append(DirN,C1,C2), name(X,C2),
   write('Reading '), write(X), nl,
   reado(X), 
   write('That was '), write(X), nl,
   get0(_),
   fail.
read_test.

% The possible ontologies

ontol_directory("D:/cmellish/Files/ontologies/").

ontol('earthrea.owl').
ontol('fusion.owl').
ontol('wine.owl').
ontol('food.owl').
ontol('pizza.owl').
ontol('newpizza.owl').
ontol('oldpizza.owl').
ontol('http___loki.cae.drexel.edu_~wbs_ontology_2004_08_fgdc-csdgm.owl.txt').
ontol('http___mged.sourceforge.net_ontologies_MGEDOntology.owl.txt').
ontol('http___xobjects.seu.edu.cn_resource_drm_orel_orel0_5.owl#.txt').
ontol('http___www.biopax.org_Downloads_Level2v0.94_biopax-level2.owl.txt').
ontol('DOLCE-Lite.owl').
ontol('univ-bench.owl').

newaxiom(X) :- assert(axiom(X)).

r(X,Y,Z) :- rdf_global_id(X,X1), rdf_global_id(Y,Y1), rdf_global_id(Z,Z1),
   rdf(X1,Y1,Z1), % write(X1-Y1-Z1), 
   rdf_retractall(X1,Y1,Z1).

non_anon(X) :- \+ (atom(X),name(X,[95|_])).

% deal1 - deal with classes

deal1 :- r(X,rdf:type,owl:'Ontology'),
        rdf_retractall(X,_,_),
        fail.
deal1 :- rdf_retractall(_,rdfs:label,_),
        rdf_retractall(_,rdf:type,owl:'Class'),
        rdf_retractall(_,rdf:type,owl:'ObjectProperty'),
        rdf_retractall(_,rdf:type,owl:'DatatypeProperty'),
        rdf_retractall(_,rdfs:comment,_),
        rdf_retractall(_,rdf:datatype,_),
        fail.

deal1 :- r(X,rdfs:subClassOf,Y),
        follow(X,X1),
        follow(Y,Y1),
        newaxiom(subclass_of(X1,Y1)), fail.
deal1 :- r(X,owl:disjointWith,Y),
        follow(X,X1),
        follow(Y,Y1),
        newaxiom(disjoint_classes([X1,Y1])), fail.
deal1 :- r(X,owl:equivalentClass,Y),
        follow(X,X1),
        follow(Y,Y1),
        newaxiom(equivalent_classes([X1,Y1])), fail.
deal1 :- r(X,rdfs:subPropertyOf,Y),
        follow(X,X1),
        follow(Y,Y1),
        newaxiom(subproperty_of(X1,Y1)), fail.
deal1 :- r(X,owl:equivalentProperty,Y),
        follow(X,X1),
        follow(Y,Y1),
        newaxiom(equivalent_properties([X1,Y1])), fail.
deal1 :- r(X,owl:inverseOf,Y),
        follow(X,X1),
        follow(Y,Y1),
        newaxiom(object_property(X1,[inverse_of(Y1)])), fail.
deal1 :- r(X,rdfs:domain,Y),
        follow(X,X1),
        follow(Y,Y1),
        newaxiom(object_property(X1,[domain(Y1)])), fail.
deal1 :- r(X,rdfs:range,Y),
        follow(X,X1),
        follow(Y,Y1),
        newaxiom(object_property(X1,[range(Y1)])), fail.

deal1 :- rdf(X,owl:unionOf,L), non_anon(X),
   r(X,owl:unionOf,L), 
   follow_list(L,L1),
   follow(X,X1),
   newaxiom(equivalent_classes([X1,union_of(L1)])), fail.
deal1 :- rdf(X,owl:intersectionOf,L), non_anon(X),
   r(X,owl:intersectionOf,L), 
   follow_list(L,L1),
   follow(X,X1),
   newaxiom(equivalent_classes([X1,intersection_of(L1)])), fail.
deal1 :- rdf(X,owl:oneOf,L), non_anon(X),
   r(X,owl:oneOf,L), 
   follow_list(L,L1),
   follow(X,X1),
   newaxiom(equivalent_classes([X1,one_of(L1)])), fail.
deal1 :- rdf(X,owl:complementOf,Y), non_anon(X),
   r(X,owl:complementOf,Y), 
   follow(Y,Y1),
   follow(X,X1),
   newaxiom(equivalent_classes([X1,complement_of(Y1)])), fail.

% Interesting class types

deal1 :- r(X,rdf:type,owl:'TransitiveProperty'),
   follow(X,X1),
   newaxiom(object_property(X1,[transitive])),
   assert(transitive(X1)), fail.

deal1 :- r(X,(rdf:type),(owl:'FunctionalProperty')),
   follow(X,X1),
   newaxiom(object_property(X1,[functional])), fail.

deal1 :- r(X,(rdf:type),(owl:'InverseFunctionalProperty')),
   follow(X,X1),
   newaxiom(object_property(X1,[inverse_functional])), fail.

% OWL properties of individuals (apart from type)

deal1 :- r(X,rdf:type,owl:'AllDifferent'),
         r(X,owl:distinctMembers,List),
         follow_list(List,L),
         newaxiom(all_different(L)), fail.

deal1 :- r(X,owl:sameAs,Y),
         follow(X,X1),
         follow(Y,Y1),
         newaxiom(equivalent_classes([one_of([X1]),one_of([Y1])])), fail.


deal1 :- r(X,owl:differentFrom,Y),
         follow(X,X1),
         follow(Y,Y1),
         newaxiom(equivalent_classes(['BOTTOM',intersection_of([one_of([X1]),one_of([Y1])])])),
         fail.


% Boring types

deal1 :- class_type(Y), rdf_global_id(Y,Y1), r(_,rdf:type,Y1), fail.

deal1.


% deal2 - deal with instances of concepts
%
% Note that instances and their classes can be anonymous

% Interesting types

deal2 :- rdf(X,rdf:type,Y), non_anon(X), r(X,rdf:type,Y), 
   sname(X,X1),
   follow(Y,Y1),
   newaxiom(instance_of(X1,Y1)), fail.
deal2.

% deal3 - deal with other RDF triples among individuals

deal3 :- rdf(X,P,Y), non_anon(X), 
   r(X,P,Y), 
   (rdf_global_id(owl:_,P) -> (write('*** OWL property missed: '),write(X-P-Y),nl /*,break*/ ); true),
   sname(X,X1),
   follow(Y,Y1),
   sname(P,P1),
   newaxiom(related(X1,P1,Y1)), fail.

deal3.

% sname - short name for an individual

sname(X:Y,X:Y1) :- xmlify(Y,Y1), !.
sname(literal(type(_,I)),I1) :- !,
   sname(literal(I),I1).
sname(literal(lang(_,I)),X) :- !,
   sname(literal(I),X).
sname(literal(I),I2) :- !,
   name(I,Na), append(Na,[124],End), name(I1,[124|End]), xmlify(I1,I2).
sname(N,Na) :-
   rdf_split_url(_,Na1,N), xmlify(Na1,Na), !.

% follow - name of (possibly complex) class
%
% don't delete links from anonymous classes, because they may be reused

follow(X:Y,X:Y) :- !.
follow(literal(type(_,I)),I1) :- !,
   follow(literal(I),I1).
follow(literal(lang(_,I)),X) :- !,
   follow(literal(I),X).
follow(literal(I),I1) :- !,
   name(I,Na), append(Na,[124],End), name(I1,[124|End]).
follow(X,Na) :-  
   % non-anonymous thing
   rdf_global_id(X1,X),
    (X1=_:N -> (X1=_:N); X1=N), non_anon(N), 
   rdf_split_url(_,Na,N), !.
follow(X,restriction(P1,[R])) :-
   rdf(X,rdf:type,owl:'Restriction'), !,
   rdf(X,owl:onProperty,P),
   follow(P,P1),
   follow_restr(X,R).
follow(X,union_of(List)) :- 
   rdf(X,owl:unionOf,Y), !,
   follow_list(Y,List).
follow(X,intersection_of(List)) :- 
   rdf(X,owl:intersectionOf,Y), !,
   follow_list(Y,List).
follow(X,complement_of(X1)) :- 
   rdf(X,owl:complementOf,Y), !,
   follow(Y,X1).
follow(X,one_of(List)) :-
   rdf(X,owl:oneOf,Y), !,
   follow_list(Y,List).
follow(X,Y2) :-
   rdf(X,rdf:datatype,Y), Y=literal(Y1), !,
   follow(Y1,Y2).
follow(X,_) :-
   write('***UNABLE TO FOLLOW FROM: '), write(X), nl, break, fail.

follow_list(X,[X1|L]) :-
   rdf(X,rdf:first,Y), !,
   (r(X,rdf:type,rdf:'List'); true), !,
   follow(Y,X1),
   rdf(X,rdf:rest,Z),
   follow_list(Z,L).
follow_list(_,[]).

follow_restr(X,all_values_from(R)) :-
   rdf(X,owl:allValuesFrom,Y), !,
   follow(Y,R).
follow_restr(X,some_values_from(R)) :-
   rdf(X,owl:someValuesFrom,Y), !,
   follow(Y,R).
follow_restr(X,cardinality(R)) :-
   rdf(X,owl:cardinality,literal(I)), !,
   atom_codes(I,Co), number_codes(R,Co).
follow_restr(X,min_cardinality(R)) :-
   rdf(X,owl:minCardinality,literal(I)), !,
   atom_codes(I,Co), number_codes(R,Co).
follow_restr(X,max_cardinality(R)) :-
   rdf(X,owl:maxCardinality,literal(I)), !,
   atom_codes(I,Co), number_codes(R,Co).   
follow_restr(X,value(R)) :-
   rdf(X,owl:hasValue,Y), !,
   follow(Y,R).

% Make a name suitable to appear in XML

xmlify(X,Y) :- name(X,X1), xml1(X1,Y1), name(Y,Y1).

xml1([],[]) :- !.
xml1([60|L],[38,108,116,59|L1]) :- !, xml1(L,L1).    % < -> &lt;
xml1([62|L],[38,103,116,59|L1]) :- !, xml1(L,L1).    % > -> &gt;
xml1([38|L],[38,97,109,112,59|L1]) :- !, xml1(L,L1). % & -> &amp;
xml1([39|L],[38,97,112,111,115|L1]) :- !, xml1(L,L1).    % ' -> &apos;
xml1([34|L],[38,113,117,111,116,59|L1]) :- !, xml1(L,L1).    % " -> &quot;
xml1([X|L],[X|L1]) :- xml1(L,L1).

% Special types with nothing special to say

class_type(owl:'Class').
class_type(owl:'ObjectProperty').
class_type(owl:'DatatypeProperty').
class_type(owl:'Ontology').
class_type(owl:'DataRange').
class_type(rdfs:'Class').
class_type(rdf:'Property').


% Properties that don't indicate relation(_,_,_) axioms

interesting_property(rdfs:subClassOf).
interesting_property(owl:disjointWith).
interesting_property(owl:equivalentClass).
interesting_property(rdfs:subPropertyOf).
interesting_property(owl:equivalentProperty).
interesting_property(owl:inverseOf).
interesting_property(rdfs:domain).
interesting_property(rdfs:range).
interesting_property(rdf:type).
interesting_property(owl:onProperty).
interesting_property(owl:unionOf).
interesting_property(owl:intersectionOf).
interesting_property(owl:complementOf).
interesting_property(owl:oneOf).
%interesting_property(rdf:first). These can be used inside instance data
%interesting_property(rdf:rest).
interesting_property(owl:allValuesFrom).
interesting_property(owl:someValuesFrom).
interesting_property(owl:cardinality).
interesting_property(owl:minCardinality).
interesting_property(owl:maxCardinality).
interesting_property(owl:hasValue).
interesting_property(rdfs:label).
interesting_property(rdfs:comment). 
interesting_property(owl:imports).
interesting_property(rdf:datatype).




