% The core of realisation:
%
% real(+Formula,-Tree)
%
% Return a possible realisation for a formula. In general, returns
% several possibilities on backtracking

%old vs new approach to A
oldfor :- fail.

real(subclass_of(C1,restriction(R,L)),                % card 1, plus all_values_from
   s(C1):[np:[det:[the],nbar:[N,pp:[of,NP1]]],VP]) :-
   select(all_values_from(D),L,[cardinality(1)]),
   nRoleDescrip(R,N),
   npClassDescrip(C1,pos,NP1), vpClassDescrip(D,VP).

%real(subclass_of(C1,C2),
%  s(C1):[D1,vp:[verb:[be],np:[det:[indef],nbar:[n:[kind],pp:[prep:[of],D2]]]]]) :- 
%   npClassDescrip(C1,pos,D1), nbarClassDescrip(C2,D2). 

real(subclass_of(C1,restriction(R,[all_values_from(C2)])),
   s(of(R,C1)):[np:[det:[any],nbar:NB1],VP]) :-
   \+ oldfor,
   vc(R,verb:[have],nbar:NB), !, 
   append(NB,[pp:[prep:[of],NP]],NB1),
   npClassDescrip(C1,pos,NP),
   vpClassDescrip(C2,VP).
real(subclass_of(C1,restriction(R,[all_values_from(C2)])),
   s(of(R,C1)):[np:[det:[any],nbar:NB1],VP]) :-
   \+ oldfor,
   vc(R,V,nbar:NB), !, 
   append(NB,[relc:[relpron:[],s:[NP,vp:[V]]]],NB1),
   npClassDescrip(C1,pos,NP),
   vpClassDescrip(C2,VP).

real(subclass_of(C1,intersection_of([C2])),Out) :- !,
   real(subclass_of(C1,C2),Out).

% other intersections done by aggregating VPs

real(subclass_of(C1,union_of([C2])),Out) :- !,
   real(subclass_of(C1,C2),Out).

real(subclass_of(C1,union_of([C2|C3s])),
   s(C1):[D1,vp:[VP1,conjn:[or],VP2]]) :-
   real(subclass_of(C1,C2),s(C1):[D1,VP1]),
   real(subclass_of(C1,union_of(C3s)),s(C1):[D1,VP2]).

real(subclass_of(C1,C2),                
     s(C1):[NP1,VP]) :-
   npClassDescrip(C1,pos,NP1),
   vpClassDescrip(C2,VP).

real(disjoint_classes([C1,C2]),
   s(C1):[D1,VP1]) :- 
   npClassDescrip(C1,pos,D1), 
   vpClassDescrip(C2,VP),
   insert_adv([never],VP,VP1).

real(equivalent_classes([C1,C2]),s(C1):[D1,vp:[verb:[be],D2]]) :-
   npClassDescrip(C1,pos,D1), npClassDescrip(C2,pos,D2). 

real(subproperty_of(P1,P2),s(P1):[D1,vp:[verb:[be],np:[det:[indef],nbar:[n:[kind],pp:[prep:[of],D2]]]]]) :-
   npRoleDescrip(P1,D1), npRoleDescrip(P2,D2). 

real(object_property(P,[domain(D)]),s(D):[np:[adv:[only],NP1],vp:[aux:[can],vp:[V,NP]]]) :-
   npClassDescrip(D,pos,NP1), vc(P,V,NB), 
   NP=np:[det:[indef],NB]. 

% ***the following not yet done
real(object_property(P,[range(D)]),s(P):[every,N,something,V,is,NP]) :-
   vc(P,V,N), npClassDescrip(D,pos,NP). 
real(object_property(P,[functional]),s(P):[NP1,always,exists,and,is,unique]) :-
   nRoleDescrip(P,NP1). 
real(object_property(P,[inverse_of(D)]),s(P):[NP1,is,the,opposite,of,NP2]) :-
   npRoleDescrip(P,NP1), npRoleDescrip(D,NP2). 
real(equivalent_properties([P1,P2]),s(P1):[D1,is,the,same,as,D2]) :-
   npRoleDescrip(P1,D1), npRoleDescrip(P2,D2). 

% npClassDescrip(Class,Pos,NP) 
%
% Build an NP realising a class expression. Pos indicates how disjunction is to be 
% realised - pos means as "or"; neg means as "and". Conjunction will be realised
% as the possibility not used for disjunction.
%
% This always builds a singular NP, using indefinite determiners.

% unions - build a conjoined NP
npClassDescrip(union_of(L),_,NP) :- !, build_disj_np(L,NP).

% single intersection
npClassDescrip(intersection_of([C]),_,NP) :- !,
   npClassDescrip(C,_,NP).

% not on its own
npClassDescrip(complement_of(C),_,np:[det:[indef],nbar:[n:[thing],relc:[relpron:[which],
   s:[np:[trace:[]],VP]]]]) :- !, 
   vpClassDescrip(complement_of(C),VP).

% intersection with not
npClassDescrip(intersection_of(List),pos,np:[NP2,conjn:[but,not],NP1]) :-
   select(complement_of(C1),List,Cs), !, \+ Cs=[],
   npClassDescrip(C1,neg,NP1),
   npClassDescrip(intersection_of(Cs),neg,NP2).

% intersections - try conjoined NPs if more than one real noun
% (otherwise will build a complex single NP)
npClassDescrip(intersection_of(L),_,NP) :-
   %conjs_of(intersection_of(L),Mods), Mods = [_,_|_],
   %select(N,Mods,Mods1),
   %class_name(N), lex(N,noun,_), 
   %select(N1,Mods1,_), class_name(N1), lex(N1,noun,_), !,
   build_conj_nps(L,NP).

% one of

npClassDescrip(one_of([X]),_,INP) :- 
   indiv_np(X,INP), !.
npClassDescrip(one_of(List),_,
   np:[det:[one],nbar:[n:[],pp:[prep:[of],NPs]]]) :- 
   real_indiv_list(List,NPs), !.

npClassDescrip('TOP',_,np:[det:[any],nbar:[n:[thing]]]) :- !.

npClassDescrip('BOTTOM',_,np:[det:[],nbar:[n:[nothing]]]) :- !.

npClassDescrip(C,_,np:[det:[indef],nbar:[n:[N]]]) :-
   class_name(C), lex(C,noun,N), !.

npClassDescrip(C,_,np:[det:[indef],nbar:[n:[thing],adj:[A]]]) :- 
   class_name(C), lex(C,adj,A).

npClassDescrip(C,_,np:[det:[],nbar:[n:[],adj:[A]]]) :- 
   class_name(C), lex(C,adj,A), !.

npClassDescrip(C,_,np:[det:[indef],nbar:[n:[thing],
      pp:[prep:[in],[np:[det:[],nbar:[n:[class,N]]]]]]]) :- 
   class_name(C), lex(C,_,N), !.

% intersection or other complex
% Build a complex NP 
% First select something that will be the noun (or use "thing"), then
% find adjectives to go before it and other modifiers to go after it

% only one real noun present
npClassDescrip(C,_,np:[det:[indef],nbar:[Ad,nbar:[n:[No]|ModTs]]]) :-
   conjs_of(C,Mods), 
   select(N,Mods,Mods1),
   class_name(N), lex(N,noun,No),
   \+ ((member(Z,Mods1),class_name(Z),lex(Z,noun,_))),  !,
   % don't allow Ns as modifiers (must conjoin the NPs) if there is a real noun
   moddescrip(intersection_of(Mods1),_,List),
   select_adjs(List,Ad,ModTs).

% no real nouns present
npClassDescrip(C,_,np:[det:[indef],nbar:[Ad,nbar:[n:[thing]|Ms]]]) :-
   conjs_of(C,Mods), 
   \+ ((member(Z,Mods),class_name(Z),lex(Z,noun,_))),  !,
   moddescrip(intersection_of(Mods),_,List),
   select_adjs(List,Ad,Ms).

% Realise an individual

indiv_np(X,np:[det:[],nbar:[n:[N]]]) :-
   lex(X,name,N), !.
indiv_np(X,np:[det:[],nbar:[adj:[A],nbar:[n:[]]]]) :-
   lex(X,adj,A), !.
indiv_np(X,np:[det:[the],nbar:[n:[individual],n:[N]]]) :-
   lex(X,_,N).

% Realise a list of individuals. Unless all are names, need
% to qualify the list

real_indiv_list(List,NPs) :-
   ((member(X,List),\+ lex(X,name,_)) -> 
      NPs=np:[det:[the],nbar:[n:[individuals],NP1]];
      NPs = NP1),
   real_indiv_list1(List,NP1).

   real_indiv_list1([X],np:[det:[],nbar:[n:[N]]]) :- !,
      lex(X,_,N).
   real_indiv_list1([X1,X2],np:[NP1,conjn:[or],NP2]) :- !,
      real_indiv_list1([X1],NP1), real_indiv_list1([X2],NP2).
   real_indiv_list1([X|Xs],np:[NP1,conjn:[','],NP2]) :-
      real_indiv_list1([X],NP1), real_indiv_list1(Xs,NP2).

% Build a conjunction of NPs for a list of concepts
% Doesn't work for disjunctions, because it packs multiple
% juncts into a single NP

build_conj_nps(List,NP) :-
   findall(N,(member(N,List),class_name(N),lex(N,noun,_)),Nouns),
   findall(N,(member(N,List),\+ member(N,Nouns)),NonNouns),
   build_conj_nps1(Nouns,NonNouns,NP).

   % 2 or more nouns - put modifiers on last one
   build_conj_nps1([N1,N2|RN],NonNouns,NP) :- !,
      append(First,[Last],[N1,N2|RN]), !,
      build_nnps(First,NPs),
      npClassDescrip(intersection_of([Last|NonNouns]),neg,NPL),
      append(NPs,[NPL],AllNPs),
      conjoin_phrases(AllNPs,neg,NP).
   % 1 noun - create a second "thing" NP 
   build_conj_nps1([N],NonNouns,NP) :- !,
      npClassDescrip(N,pos,NP1),
      npClassDescrip(intersection_of(NonNouns),pos,NP2),
      conjoin_phrases([NP1,NP2],neg,NP).
   % no nouns - create 2 "thing" NPs
   build_conj_nps1([],[NN|NonNouns],NP) :-
      npClassDescrip(NN,pos,NP1),
      npClassDescrip(intersection_of(NonNouns),pos,NP2),
      conjoin_phrases([NP1,NP2],neg,NP).      
   
   build_nnps([],[]) :- !.
   build_nnps([C|Cs],[NP|NPs]) :-
      npClassDescrip(C,pos,NP),
      build_nnps(Cs,NPs).

% build a disjunctive NP

build_disj_np(List,NP) :-
   build_nnps(List,NPs),
   conjoin_phrases(NPs,pos,NP).

% Build a class description as an NP with an empty determiner. Just
% builds an NP and strips off the determiner

nbarClassDescrip(C,np:[det:[]|L]) :- npClassDescrip(C,_,np:[det:_|L]).

% card(+List,-Det,?Rest)

% Find cardinality information in a restriction list, List, and
% return Words to put in a determiner and the rest of the list.
% Returns a simple indefinite if there is no explicit card restriction
% Also removes any some_values_from('TOP') from the list

card(L,Det,L3) :-
   card1(L,Det,L1),
   remove_exists_top(L1,L2), L2=L3.

   card1(List,det:[at,least,Num|Cs],Rest1) :-
      select(min_cardinality(Num),List,Rest), !,
      restcard(Rest,Cs,Rest1).
   card1(List,det:[at,most,Num|Cs],Rest1) :-
      select(max_cardinality(Num),List,Rest), !,
      restcard(Rest,Cs,Rest1).
   card1(List,det:[no|Cs],Rest1) :- 
      select(cardinality(0),List,Rest), !,
      restcard(Rest,Cs,Rest1).
   card1(List,det:[exactly,Num|Cs],Rest1) :-
      select(cardinality(Num),List,Rest), !,
      restcard(Rest,Cs,Rest1).
   card1(List,det:[indef],List).

   restcard(Rest,[and|Cs],Rest1) :-
      card1(Rest,X,_), \+ (X=det:[indef]), !,
      card1(Rest,det:Cs,Rest1).
   restcard(Rest,[],Rest).

   remove_exists_top(L,L2) :-
      select(some_values_from('TOP'),L,L1), !,
      remove_exists_top(L1,L2).
   remove_exists_top(L,L).

% class description as a set of modifiers for an NP
%
% e.g. as could be used in the context "An X ..."

moddescrip(C,Spec,[WholeMod]) :-
   moddescrip1(C,Spec,M),
   order_mods(M,Mods),
   agg_mods(Mods,Mods1),
   conjoin_phrases(Mods1,neg,WholeMod).

moddescrip1(intersection_of(L),post,CM) :- !,             % Set of modifiers
   build_mods(L,post,CM).
moddescrip1(C,pre,[adj:[A]]) :-                           % C - adjective
   class_name(C), lex(C,adj,A), !.
moddescrip1(restriction(R,[value(A)]),pre,[adj:[Ad]]) :-    % adjective for role values
   vc(R,verb:[have],_),
   lex(A,adj,Ad), !.
moddescrip1(C,post,[pp:[prep:[with],NP]]) :-                 % with NP
   vpClassDescrip(C,VP),
   VP=vp:[verb:[have],NP].
moddescrip1(C,post,[pp:[prep:[in],[np:[det:[],nbar:[n:[class,C]]]]]]) :-  % in class C
   class_name(C), \+ lex(C,noun,_), 
   \+ lex(C,adj,_), !.
moddescrip1(restriction(R,[some_values_from(D)]),post,                  % Prep NP
   [pp:[prep:[P],np:[det:[at,least,one],N]]]) :-
   lex(R,prep,P),
   nRoleDescrip(D,N).
moddescrip1(restriction(R,[all_values_from(D)]),post,      % any N of which ...
	 [relc:[relpron:[','],s:[np:[det:[any],nbar:NB1],VP]]]) :-
   \+ oldfor,
   vc(R,verb:[have],nbar:NB),
   append(NB,[pp:[prep:[of],np:[det:[which],nbar:[n:[]]]]],NB1),
   vpClassDescrip(D,VP).
moddescrip1(restriction(R,[all_values_from(D)]),post,      % for which every Role ...
	 [relc:[relpron:[',',for,which],s:[np:[det:[every],NB],VP]]]) :-
   oldfor,
   vc(R,verb:[have],NB),
   vpClassDescrip(D,VP).
moddescrip1(restriction(R,[all_values_from(D)]),post,      % where everything it Roles...
	 [relc:['X'(Num),relpron:[where],s:[np:[det:[],nbar:[n:[everything],
           relc:[relpron:[],s:[np:[det:[],nbar:[n:[pron(Num)]]],
			     vp:[V,np:[trace:[]]]]]]],VP]]]) :-
   oldfor,
   vc(R,V,_), \+ (V=(verb:[have])),
   vpClassDescrip(D,VP).
moddescrip1(Mod,post,[relc:[relpron:[which],s:[np:[trace:[]],VP]]]) :-   % default - subject relative clause
   (oldfor -> \+ Mod=[restriction(_,[all_values_from(_)])]; true),           % ("can only" VP doesn't work)
   vpClassDescrip(Mod,VP).

   % build a list of trees for a set of concepts realised as modifiers
   % second argument specifies pre/post/_

   build_mods([],_,[]).
   build_mods([C|Cs],Spec,[ML1|ML2]) :- moddescrip(C,Spec,[ML1]), build_mods(Cs,Spec,ML2).

   % select the adjectives from a list of modifiers
   select_adjs(L,adj:A3,L2) :-
      select(adj:A1,L,L1), !,
      select_adjs(L1,adj:A2,L2),
      append(A1,A2,A3).
   select_adjs(L,adj:[],L).
 
   % aggregate a list of (post-) modifiers into a shorter list of more
   % complex modifiers. This could be made much more general.

   agg_mods([M],[M]) :- !.
   agg_mods([],[]) :- !.
   agg_mods(L,L3) :-
      select(M1,L,L1),
      select(M2,L1,L2),
      agg(M1,M2,M3), !,
      agg_mods([M3|L2],L3).
   %agg_mods(L,[L1]) :- conjoin_phrases(L,pos,L1).
   agg_mods(L,L).

   % aggregate two modifiers into one. Can only do this if they have
   % the same category (C) and some shared non-empty initial list of
   % daughters (Pre)
   agg(C:[Pre,C1:L1],C:[Pre,C1:L2],C:[Pre,C1:[C1:L1,conjn:[and],C1:L2]]).

   % order a set of post-modifiers: PPs before relative clauses
   order_mods(Ms,[pp:M|M2s]) :-
      select(pp:M,Ms,M1s), !,
      order_mods(M1s,M2s).
   order_mods(Ms,Ms).

% NP descriptions for roles
% Builds a singular indefinite NP for the value of a role

npRoleDescrip(R,np:[det:[indef],NB]) :-
   vc(R,verb:[have],NB), !.

% Build an Nbar 
nRoleDescrip(C,nbar:L) :- 
   npRoleDescrip(C,np:[det:_,nbar:L]).

% Build a VP from a concept

vpClassDescrip(restriction(R,L),vp:[V,np:[Words,NB]]) :-    % cardinality only
   vc(R,V,NB), 
   card(L,Words,[]), !.
vpClassDescrip(restriction(R,[some_values_from(C2)]),vp:[V,NP]) :-       % some values only (N not given) 
   vc(R,V,NB), NB=nbar:[n:[thing]|_], !,
   npClassDescrip(C2,pos,NP).
vpClassDescrip(restriction(R,L),vp:[V,np:[det:Words,nbar:NB1]]) :-             % some values + (card or N)
   vc(R,V,nbar:NB),
   card(L,det:Words,[some_values_from(C2)]), !,
   ((Words=[indef]; last(Words,1)) -> 
       (Pron=[which],NP=np:[trace:[]]); 
       (Pron=[],NP=np:[det:[at,least,one],nbar:[n:[],pp:[prep:[of],
                          np:[det:[],nbar:[n:[which]]]]]])), !,
   append(NB,[relc:[relpron:Pron,s:[NP,VP]]],NB1),
   vpClassDescrip(C2,VP).
vpClassDescrip(restriction(R,[all_values_from(C)]),VP) :-                        % all values only (have)
   \+ oldfor,
   vc(R,verb:[have],NB),
   VP=vp:[verb:[need],s:[np:[det:[any],NB],vp:[aux:[to],VP1]]],
   vpClassDescrip(C,VP1).
vpClassDescrip(restriction(R,[all_values_from(C)]),VP) :-                        % all values only (have)
   \+ oldfor,
   vc(R,V,nbar:NB),
   VP=vp:['X'(Num),verb:[need],s:[np:[det:[any],nbar:NB1],vp:[aux:[to],VP1]]],
   append(NB,[relc:[relpron:[],s:[np:[det:[],nbar:[n:[pron(Num)]]],vp:[V,np:[trace:[]]]]]],NB1),
   vpClassDescrip(C,VP1).
vpClassDescrip(restriction(R,L),vp:[V,np:[det:Words,nbar:NB1]]) :-             % all values + card
   oldfor,
   vc(R,V,nbar:NB),
   card(L,det:Words,[all_values_from(C2)]), 
   \+ (Words=[indef]), !,
   ((Words=[indef]; last(Words,1)) -> Pron=[which]; Pron=[each,of,which]), !,
   append(NB,[relc:[relpron:Pron,s:[np:[trace:[]]],VP]],NB1),
   vpClassDescrip(C2,VP1),
   insert_aux([must],VP1,VP).
vpClassDescrip(restriction(R,[all_values_from(C)]),VP3) :-              % all values only (N not given)
   oldfor,
   vc(R,V,NB), NB=nbar:[n:[thing]|_], !,
   VP1=vp:[V,NP],
   npClassDescrip(C,pos,NP),
   insert_adv([only],VP1,VP3).
   % insert_aux([can],VP2,VP3). Does not work well in embedded contexts
vpClassDescrip(restriction(R,[all_values_from(C)]),VP3) :- !,            % all values only (N given)
   oldfor,
   vc(R,V,nbar:NB),
   VP1=vp:[V,np:[det:[indef],nbar:NB1]],
   append(NB,Mod,NB1),
   moddescrip(C,post,Mod),
   insert_adv([only],VP1,VP3).
   % insert_aux([can],VP2,VP3).  Does not work well in embedded contexts
vpClassDescrip(restriction(R,L),vp:[V,np:[det:Words,nbar:NB1]]) :-                  % value + card
   vc(R,V,nbar:NB),
   card(L,det:Words,[value(X)]), 
   \+ (Words=[indef]), !,
   indiv_np(X,INP),
   append(NB,[relc:[relpron:[which],s:[np:[trace:[]],vp:[verb:[include],INP]]]],NB1).
vpClassDescrip(restriction(R,[value(X)]),                             % has Name as a Noun
	 vp:[verb:[have],INP,pp:[prep:[as],np:[det:[indef],NB]]]) :-
   indiv_np(X,INP),
   vc(R,verb:[have],NB).
vpClassDescrip(restriction(R,[value(X)]),                             % Roles Name
	 vp:[V,INP]) :- !,
   vc(R,V,_), \+ (V=(verb:[have])),
   indiv_np(X,INP).
vpClassDescrip(restriction(R,Res),Out) :- !,                     % more complex restrictions
   Res=[_,_|_],
   findall(restriction(R,[RS]),member(RS,Res),Restrs),
   vpClassDescrip(intersection_of(Restrs),Out).
vpClassDescrip(intersection_of(L),VP) :-                        % conjunction of VPs
   \+ (member(X,L),class_name(X),lex(X,noun,_)),               % if there are nouns present, complex NP will be better
   build_vps(L,VL),
   aggregate_vps(VL,VL1),
   conjoin_phrases(VL1,neg,VP).   

vpClassDescrip(complement_of(restriction(R,[min_cardinality(N)])),VP) :- N>0, !,
   N1 is N-1,
   vpClassDescrip(restriction(R,[max_cardinality(N1)]),VP).
vpClassDescrip(complement_of(restriction(R,[max_cardinality(N)])),VP) :- !,
   N1 is N+1,
   vpClassDescrip(restriction(R,[min_cardinality(N1)]),VP).
vpClassDescrip(complement_of(restriction(R,[all_values_from(C)])),VP) :- !,
   vpClassDescrip(restriction(R,[some_values_from(complement_of(C))]),VP).
vpClassDescrip(complement_of(restriction(R,[X,Y|L])),VP) :- !, 
   findall(complement_of(restriction(R,[Re])),member(Re,[X,Y|L]),Restrs),
   vpClassDescrip(union_of(Restrs),VP).
vpClassDescrip(complement_of(intersection_of(L)),VP) :- !,
   findall(complement_of(X),member(X,L),L1),
   vpClassDescrip(union_of(L1),VP).
vpClassDescrip(union_of(intersection_of(L)),VP) :- !,
   findall(complement_of(X),member(X,L),L1),
   vpClassDescrip(intersection_of(L1),VP).
vpClassDescrip(complement_of(complement_of(C)),VP) :- !,
   vpClassDescrip(C,VP).
vpClassDescrip(complement_of(C),VP) :- !,
   vpClassDescrip(C,VP1),
   negate(VP1,VP).


vpClassDescrip(C,vp:[verb:[be],NP]) :-                        % other or complex concept
   npClassDescrip(C,pos,NP).

% find a verb and complement for roles, according to syntactic category

vc(R,verb:[V],nbar:[n:[thing]]) :- lex(R,verb,V), !.
vc(R,verb:[have],nbar:[n:[N]]) :- lex(R,noun,N), !.
vc(R,verb:[be,P],nbar:[n:[thing]]) :- lex(R,prep,P), !.
vc(R,verb:[V],nbar:[n:[R1]]) :- lex(R,vn,V+R1), !.
vc(R,verb:[V,P],nbar:[n:[thing]]) :- lex(R,vp,V+P), !.
vc(R,verb:[have],nbar:[n:[value],
   pp:[prep:[for],np:[det:[the],nbar:[n:[role],n:[U]]]]]) :- lex(R,unknown,U).

% Build a list of VPs for a list of concepts

build_vps([],[]) :- !.
build_vps([C|Cs],[VP|VPs]) :- vpClassDescrip(C,VP), \+ (VP=_:[_,conjn:_,_]), build_vps(Cs,VPs).

% Aggregate a list of VPs into a shorter list if possible

aggregate_vps(VL,VL3) :-
   append(First,[vp:VP1|Rest],VL),
   select(vp:VP2,Rest,Rest1),
   agg_vp(VP1,VP2,VP3), \+ VP3=[_,conjn:_,_], !,
   append(First,Rest1,AllRest),
   aggregate_vps([vp:VP3|AllRest],VL3).
aggregate_vps(VPs,VPs).

   agg_vp([aux:A,vp:VP1],[aux:A,vp:VP2],[aux:A,vp:VP3]) :- !,
      agg_vp(VP1,VP2,VP3).
   agg_vp([adv:A,vp:VP1],[adv:A,vp:VP2],[adv:A,vp:VP3]) :- !,
      agg_vp(VP1,VP2,VP3).
   % if anything, the first VP is complex
   agg_vp([verb:L|Cs1],[verb:L|Cs2],[verb:L,vp:[vp:[verb:[]|Cs2],conjn:[Conj],vp:[verb:[]|Cs1]]]) :- 
      \+ Cs1=[], \+ Cs2=[], !,
      (Cs1=[vp:[_,conjn:_,_]] -> Conj=','; Conj=and).
   agg_vp(VP1,VP2,[vp:VP1,conjn:[and],vp:VP2]).

conjoin_phrases([C:P1,C:P2],neg,C:[conjn:[both],C:P1,conjn:[and],C:P2]) :- !.
conjoin_phrases([C:P1,P2|Ps],pos,C:[conjn:[either]|L]) :- !, conjoin_phrases1([C:P1,P2|Ps],pos,C:L).
conjoin_phrases(Ps,PN,P1) :- conjoin_phrases1(Ps,PN,P1).

conjoin_phrases1([X],_,X) :- !.
conjoin_phrases1([C:L1,C1:L2],Pos,C:[C:L1,conjn:[Conj],C1:L2]) :- !,
   conj(Pos,Conj).
conjoin_phrases1([C:L1|L],Pos,C:[C:L1,conjn:[','],Cs]) :- conjoin_phrases1(L,Pos,Cs).

   conj(pos,or).
   conj(neg,and).

% Class names allowed in formulae

class_name(X) :- atomic(X), !.
class_name(noun(_)) :- !.
class_name(adj(_)) :- !.
class_name(unknown(_)).

% Property names
% (this definition isn't used, but documents what is possible)

prop_name(X) :- atomic(X), !.
prop_name(prep(_)) :- !.
prop_name(verb(_)) :- !.
prop_name(vn(_,_)) :- !.
prop_name(vp(_,_)) :- !.
prop_name(unknown(_)) :- !.

% Individual names
% (again, not used)

indiv_name(X) :- atomic(X), !.
indiv_name(unknown(_)) :- !.
indiv_name(adj(_)) :- !.
indiv_name(name(_)).

% Lexicon. Prompt the user to classify names in terms of parts of speech,
% and save the results so that it is not necessary to ask again.

:- dynamic haslex/3.
:- multifile haslex/3.

lex('TOP',_,_) :- !, fail.
lex('BOTTOM',_,_) :- !, fail.
lex(noun(N),C,R) :- !, C=noun, R=N.
lex(prep(P),C,R) :- !, C=prep, R=P.
lex(adj(A),C,R) :- !, C=adj, R=A.
lex(name(N),C,R) :- !, C=name, R=N.
lex(verb(V),C,R) :- !, C=verb, R=V.
lex(vn(V,N),C,R) :- !, C=vn, R=V+N.
lex(vp(V,P),C,R) :- !, C=vp, R=V+P.
lex(unknown(X),C,R) :- !, C=unknown, R=X.
lex(Atom,X,Y) :- haslex(Atom,X1,Y1), !, X=X1, Y=Y1.
lex(Atom,X,Y) :-
   atomic(Atom), 
   nl,
   write('Category of '), write(Atom), 
   write(' ? (noun/prep/adj/name/verb/vn/vp/unknown) '), ttyflush,
   read(X1), 
   write('Root of '), write(Atom), 
   write(' ? (A/V+A) '), ttyflush,
   read(Y1),
   assert(haslex(Atom,X1,Y1)), X=X1, Y=Y1.



