% Interface for reasoning tasks
%
% system and user go to RACER/Fact
% empty goes to structural 

:- dynamic dig1p1/0.

dig1p1 :- fail.      % Whether using DIG 1.1 (ie FACT rather than Racer)
                     % This never worked

% decode KB specification k(Sys,User,Ego)

system_KB(k(KB,_,_),KB).
user_KB(k(_,KB,_),KB).
ego_KB(k(_,_,KB),KB).

% create a new KB, or renew an existing one

:- dynamic uri/2.

new_kb(KB) :-
   dig1p1, !,
   (uri(KB,Uold) -> racer_command(KB,releaseKB(Uold),_); true),
   racer_command(KB,newKB,uri(U)),
   retractall(uri(KB,_)),
   assert(uri(KB,U)),
   memoreset(KB).
new_kb(KB) :-
   racer_tell(KB,clearKB), memoreset(KB).

% introduce axioms read in (and reset ALL KBs)

addaxioms(KB) :- findall(X,axiom(X),L), racer_tell(KB,L), memoreset(_).

% introduce just vocabulary

addvocab(KB) :- 
   all_concepts(Cs),
   all_roles(Rs),
   setof(declare_class(C),member(C,Cs),L1),
   racer_tell(KB,L1),
   setof(declare_role(R),member(R,Rs),L2),
   racer_tell(KB,L2).
   % setof(declare_individual(I),X^(axiom(X),mowl_axiom(X,I,individual)),L3),
   % racer_tell(KB,L3).

% add just one axiom

addaxiom(KB,Ax) :- racer_tell(KB,Ax), memoreset(KB).

% introduce all negated concepts

addnegs(KB) :- axiom(X), mowl_class(X,C), name(C,Na), append("neg",Na,Na1),
   name(C1,Na1), racer_tell(KB,equivalent_classes([C1,complement_of(C)])), fail.
addnegs(KB) :- memoreset(KB).

% add difference between two concepts

adddiff(KB,X,Y) :- name(X,XNa), name(Y,YNa),
   append(XNa,"AndNot",XN1), append(XN1,YNa,YN1),
   name(D1,YN1), racer_tell(KB,equivalent_classes([D1,conjunction_of([X,complement_of(Y)])])),
   append(YNa,"AndNot",YN2), append(YN2,XNa,XN2),
   name(D2,XN2), racer_tell(KB,equivalent_classes([D2,conjunction_of([Y,complement_of(X)])])),
   memoreset(KB).

% add an alias

addalias(KB,Old,New) :- racer_tell(KB,equivalent_classes([Old,New])), memoreset(KB).

% abbreviated versions of common asks, which also do useful postprocessing

subsumes(_,X,X) :- !.
subsumes(empty,X,Y) :- !, t(X,Y).
%subsumes(KB,X,Y) :- simpleopt, simple(KB,X,Res), !, 
%   wax(X), write(' is '), write(Res), nl,
%   (Res='TOP' -> true; subsumes(KB,'BOTTOM',Y)).
subsumes(KB,X,Y) :- racer_ask(KB,subsumes(X,Y)).

equiv(KB,X,Y) :- subsumes(KB,X,Y), subsumes(KB,Y,X).
children(KB,X,Y) :- racer_ask(KB,children(X),Y1), 
   flatten_syns(Y1,Y2), findall(YY,(member(YY,Y2),\+temp_class(YY)),Y).
parents(KB,X,Y) :- racer_ask(KB,parents(X),Y1), 
   flatten_syns(Y1,Y2), findall(YY,(member(YY,Y2),\+temp_class(YY)),Y).
synonyms(KB,X,Y) :- racer_ask(KB,equivalents(X),Y1), flatten_syns(Y1,Y).

% all_ versions take all returned classes, don't check for temp classes

all_parents(empty,X,Y) :- !, uall_parents(X,Y).
all_parents(KB,X,Y) :- racer_ask(KB,all_parents(X),Y). 

all_children(empty,X,Y) :- !, uall_children(X,Y).
all_children(KB,X,Y) :- racer_ask(KB,all_children(X),Y).

satisfiable(KB,X) :- racer_ask(KB,satisfiable(X)).
instances(KB,C,Inst) :- racer_ask(KB,instances(C),Inst).

% temporary classes (to be left out of results) - names start with $$

temp_class(X) :- name(X,[36,36|_]).
