% using refinement to find concepts that subsume a given one
%
% Normal form: no top level conjunctions, but allow conj inside all restrictions
% Emb Restrictions: optionally contain at most one all_values_from (includes conjns)
%   - this will always be the last element of the list
% In general, for conjunctions only refine first element in list (most recent)
% (this includes restrictions)
% Conjunctions: Require that atoms selected first, then restrs (all for a property
% before any somes)
% (first selected go at end of list)

% flags for NL optimisations

:- dynamic(nameopt/0).
:- dynamic(negopt/0).
:- dynamic(univopt/0).
:- dynamic(simpleopt/0).

negopt.     % negations must be conjoined with names with a satisfiable check
univopt.    % no vacuous universal quantifiers
simpleopt.  % look for simple cases of subexpressions
nameopt.    % give system most specific names

% refine(+X,-Y,+Emb,+SAv,-NSav,+Conj,+Test,+KBs,+Rel) - X refines into a more specific Y
% Emb is the depth of embedding (0 = top level)
% Sav/NSav are the old and new numbers of operators available
% Conj is whether conjunctions are allowed yes/no/negonly
% Test is the test whether to keep going down the tree - is instantiated at the end
% KBs is the system, user and viewpoint KB names
% Rel is the relevant concepts and properties

%refine(Y,X,D,A,NA,Conj,Test,KBs,Rel) :-
%    writeq(refine(Y,X,D,A,NA,Conj,Test,KBs,Rel)), nl, break, fail.
refine('TOP',X,D,A,NA,_,Test,KBs,Rel) :- !,
   alltop(X,D,A,NA,[],Test,KBs,Rel).
refine(restriction(P,List),restriction(P,List1),D,A,NA,C,Test,KBs,Rel) :-
   D1 is D+1,
   res_refine(List,List1,D1,A,NA,C,Test,KBs,Rel).
refine(intersection_of(L),intersection_of(L1),D,A,NA,yes,Test,KBs,Rel) :- !,
   refine_conj(L,L1,D,A,NA,Test,KBs,Rel).
refine(X,intersection_of(Res),D,A,NA,yes,Test,KBs,Rel) :- !,  % If conjunctions allowed
   add_conj([X],Res,D,A,NA,Test,KBs,Rel).
refine(X,intersection_of([Compl,X]),_,A,NA,negonly,Test,KBs,Rel) :- 
   negopt,  
   % If conjunctions not allowed, only allow negs
   atomic(X),
   A1 is A-1, A1 >= 0,
   \+ too_general_conjunct_left(KBs,[X]), 
   top(0,0,0,0,0,1,Compl,_,A1,NA,first,Test,KBs,Rel),
   \+ trivial_conjunct(KBs,Compl,[X]).

   refine_conj([X|List],[X1|List],D,A,NA,Test,KBs,Rel) :-       % just refine the first (most recent)
      refine(X,X1,D,A,NA,no,Test,KBs,Rel),                      % no conjunctions embedded inside conjunction
      \+ too_specific_conjunct(KBs,X1,List).
   refine_conj(List,Res,D,A,NA,Test,KBs,Rel) :-
      A >= 1,
      add_conj(List,Res,D,A,NA,Test,KBs,Rel).

   add_conj(List,[X|List],D,A,NA,Test,KBs,Rel) :-
      A >= 1, A1 is A-1,
      \+ too_general_conjunct_left(KBs,List), 
      alltop(X,D,A1,NA,List,Test,KBs,Rel),
      \+ trivial_conjunct(KBs,X,List),
      \+ too_specific_conjunct(KBs,X,List).

/* NO LONGER NEEDED
% transitive closure of refinement

refine_star(X,Z,Emb,SAv,NSav,Conj,Test,KBs,Rel) :-
   findall(Z-IAv,refine(X,Z,Emb,SAv,IAv,Conj,Test,KBs,Rel),List),
   continue_refine_star(List,Emb,Conj,Test,KBs,Rel,Z,NSav).

   continue_refine_star(List,_,_,_,_,_,X,NS) :- member(X-NS,List).
   continue_refine_star(List,Emb,Conj,Test,KBs,Rel,Z,NSav) :-
      member(X-INS,List), 
      refine_star(X,Z,Emb,INS,NSav,Conj,Test,KBs,Rel).
*/

% refinement introducing a definitely new formula
% (not usubsuming the original)

step_refine(X,Z,A,NA,Conj,Ind,Ind2,TestName,KBs,Rel) :- 
   make_test(TestName,Y,Test),
   refine(X,Y,0,A,A1,Conj,Test,KBs,Rel), 
   (att_tracing -> (tab(Ind), write('> '), wax(Y), nl); (write(.),ttyflush)),
   Ind1 is Ind+1,
   sr1(X,Y,Z,A1,NA,Conj,Ind1,Ind2,TestName,KBs,Rel).  

   sr1(X,Y,Z,A,NA,Conj,Ind1,Ind2,TestName,KBs,Rel) :- 
      ego_KB(KBs,KB), subsumes(KB,Y,X), !,
      step_refine(Y,Z,A,NA,Conj,Ind1,Ind2,TestName,KBs,Rel).
   sr1(_,Z,Z,A,A,_,I,I,_,_,_).

% Can't add a conjunct subsumed by an existing one

too_specific_conjunct(KBs,C,Cs) :-
   optimisations,
   ego_KB(KBs,KB),
   member(C1,Cs), subsumes(KB,C1,C), !.

% Can't leave a conjunct that is too general

too_general_conjunct_left(KBs,[restriction(P,[some_values_from(C)])|L]) :-
   optimisations,
   member(restriction(P,[all_values_from(C1)]),L),
   ego_KB(KBs,KB),
   \+ subsumes(KB,C1,C), !.
too_general_conjunct_left(KBs,[C|Cs]) :-
   optimisations,
   ego_KB(KBs,KB),
   member(C1,Cs), subsumes(KB,C,C1), !.
too_general_conjunct_left(KBs,[C|_]) :-
   redundant(KBs,C).

redundant(_,'TOP') :- !.
redundant(_,restriction(_,[all_values_from('TOP')])) :- !.
redundant(KBs,intersection_of(L)) :- !,
   too_general_conjunct_left(KBs,L).
redundant(KBs,C) :-
   optimisations,  
   system_KB(KBs,KB),             
   subsumes(KB,C,'TOP'), !.
redundant(_,restriction(_,[some_values_from('TOP')])) :- !, fail.
redundant(KBs,restriction(_,[some_values_from(C)])) :- !, 
   optimisations,
   redundant(KBs,C).
redundant(KBs,restriction(_,[all_values_from(C)])) :- 
   optimisations,
   redundant(KBs,C).

% trivial conjuncts are some negations

trivial_conjunct(KBs,complement_of(C),List) :- 
   negopt,
   system_KB(KBs,KB),
   \+ ((member(C1,List), atom(C1), satisfiable(KB,intersection_of([C,C1])))). 

% refine a all_values_from
res_refine([all_values_from(X)],[all_values_from(X1)],D,A,NA,Conj,Test,KBs,Rel) :-
   refine(X,X1,D,A,NA,Conj,Test,KBs,Rel).               % all does not change conj requirements
% refine a some_values_from
res_refine([some_values_from(X)],[some_values_from(X1)],D,A,NA,_,Test,KBs,Rel) :-
   refine(X,X1,D,A,NA,yes,Test,KBs,Rel).                % conjunctions always allowed in some_values_from

% top elements
% alltop(-C,+D,+A,-NA,+Prev,+Test,+KBs,Rel) - C is a top, A and NA the available size before and after
%    Prev is the list of previous elements of the conjunction
%    D is the depth of embedding
% Used when a completely new term is being created (e.g. added to a cnjunction)
%
% Context passed to top is one of:
%  first - no previous conjuncts
%  atom(C) - last was atom C
%  comp(C) - last was complement_of(C)
%  prev(P) - last was restriction(P,...) and we will continue with P
%  next(P) - last was restriction(P,...) and we will move to a new P

% Possible values depend on last previous element of conjunction
alltop(C,D,A,NA,[restriction(P,[all_values_from(_)])|_],Test,KBs,Rel) :- !, % only allow one
   top(0,0,1,1,1,1,C,D,A,NA,prev(P),Test,KBs,Rel).
alltop(C,D,A,NA,[restriction(P,[some_values_from(_)])|_],Test,KBs,Rel) :- !,
   top(0,0,1,1,1,1,C,D,A,NA,prev(P),Test,KBs,Rel).
alltop(C,D,A,NA,[restriction(P,_)|_],Test,KBs,Rel) :- !, % assume number restr, only allow one
   top(0,0,0,0,1,1,C,D,A,NA,prev(P),Test,KBs,Rel).
alltop(C,D,A,NA,[complement_of(C1)|_],Test,KBs,Rel) :- !,
   top(0,0,0,0,0,1,C,D,A,NA,comp(C1),Test,KBs,Rel).
alltop(C,D,A,NA,[],Test,KBs,Rel) :- !,         % first conjunct - can be almost anything
   (negopt -> Neg=0; Neg=1),
   top(1,1,1,1,0,Neg,C,D,A,NA,first,Test,KBs,Rel).
alltop(C,D,A,NA,[C1|_],Test,KBs,Rel) :-         % assume last is atomic class
   top(1,1,1,1,0,1,C,D,A,NA,atom(C1),Test,KBs,Rel).

% 1. atomic class
top(1,_,_,_,_,_,C1,_,A,A,Ctxt,Test,KBs,Rel) :-
   (nameopt -> system_KB(KBs,KB); ego_KB(KBs,KB)),
   setof(C1,Test^(down_class('TOP',C1,Test,KB,Rel),check_pos_name(C1,Ctxt)),List),
   member(C1,List).
% 2. all_values_from a class, penalty D+1
top(_,1,_,_,_,_,restriction(P,[all_values_from(C)]),D,A,NA,Ctxt,Test,KBs,Rel) :-
   D1 is D+1,
   A>=D1, A1 is A-D1,
   pickp(Ctxt,Rel,P), 
   (D=0 -> (spec_props(Rel,Spec), member(P,Spec)); true),
   (univopt -> \+ ((C='BOTTOM',on_path(Test))); true),
   alltop(C,D1,A1,NA,[],Test,KBs,Rel).
   % KBs=KBs, C='TOP', A1=NA.
% 3. some_values_from a class, penalty D+1
top(_,_,1,_,_,_,restriction(P,[some_values_from('TOP')]),D,A,A1,Ctxt,Test,_,Rel) :-
   D1 is D+1,
   A>=D1, A1 is A-D1,
   pickp(Ctxt,Rel,P),
   on_path(Test).
% 4. number restriction
top(_,_,_,1,_,_,restriction(P,List),D,A,NA,Ctxt,Test,KBs,Rel) :- 
   D1 is D+1,
   A>=D1, A1 is A-D1,     % NB counts the quantifier for a conjunctive NR only once
   pickp(Ctxt,Rel,P),
   \+ transitiveish(P,[]),
   system_KB(KBs,KB),
   find_ns(A1,NA,P,KB,List,Test).
% 5. move to a (new) property
top(_,_,_,_,1,_,C,D,A,NA,prev(P),Test,KBs,Rel) :-
   top(0,1,1,1,0,0,C,D,A,NA,next(P),Test,KBs,Rel).
% 6. negated atom
top(_,_,_,_,_,1,Compl,_,A,NA,Ctxt,Test,KBs,Rel) :-
   (nameopt -> system_KB(KBs,KB); ego_KB(KBs,KB)), 
   negations, A>=1, NA is A-1,
   % Test=Cand-_, write(Compl-Cand), nl, trace,
   findall(C1,okneg(C1,Compl,Test,Rel,KB,Ctxt),List1),
   (setof(C1,member(C1,List1),List); List=[]),
   member(C1,List),
   % \+ \+ ((Compl=C1,Test=(Cand-_),satisfiable(system,Cand))), global satisfiability test
   Compl = complement_of(C1).

   okneg(C1,Compl,Test,Rel,KB,Ctxt) :- 
      Compl=complement_of(C1),
      up_class('BOTTOM',C1,Test,KB,Rel),
      check_neg_name(C1,Ctxt).

   % OWL DL (and some reasoners) don't allow cardinality constraints on transitive properties,
   % their inverses or their superproperties

   transitiveish(P,_) :- transitive(P), !.
   transitiveish(P,L) :- axiom(subproperty_of(P1,P)), \+ member(P1,L), 
      transitiveish(P1,[P1|L]), !.
   transitiveish(P,L) :- axiom(object_property(P,[inverse_of(P1)])), \+ member(P1,L), 
      transitiveish(P1,[P1|L]), !.
   transitiveish(P,L) :- axiom(object_property(P1,[inverse_of(P)])), \+ member(P1,L),
      transitiveish(P1,[P1|L]), !.

   % pick a property, given a context spec

   % (a) continue with existing one
   pickp(prev(P),_,P) :- !.
   % (b) move to a new one
   pickp(next(P),Rels,P1) :-
      rel_props(Rels,Rel),
      all_roles(List), append(_,[P|Rest],List), !, member(P1,Rest), 
      includable(P1), once(member(P1,Rel)).
   % (c) no restrictions
   pickp(_,Rels,P) :- !,
      rel_props(Rels,Rel),
      all_roles(List), member(P,List), includable(P), once(member(P,Rel)).

   % check a name in a context
   check_pos_name(C,atom(C1)) :- !, C @< C1.
   check_pos_name(_,_).

   check_neg_name(C,comp(C1)) :- !, C @< C1.
   check_neg_name(_,_).

% down_class(C,Hole,Test,KB,Rel)
% Optimised treatment of selection of an atomic class in a positive
% context

down_class(C,Hole,Test,KB,Rel) :-
   (most_general(Test) -> down_class2(C,Hole,Test,KB,Rel); down_class1(C,Hole,Test,KB,Rel)).

% for minimisation
down_class1(C,Hole,Test,KB,Rel) :-
   \+ \+ ((C=Hole, try_descendents(Test))),  
   all_children(KB,C,List),
   member(_C,List),
   \+ _C='BOTTOM',
   \+ \+((Hole=_C,on_path(Test))), 
   \+ \+ (down_class1(_C,Hole,Test,KB,Rel)), !, % if a descendent returns, don't return C
   member(C1,List),
   \+ C1='BOTTOM',
   \+ \+((Hole=C1,on_path(Test))), 
   down_class1(C1,Hole,Test,KB,Rel).
down_class1(C,C,Test,_,Rels) :- includable(C), rel_classes(Rels,Rel), member(C,Rel), return_solution(Test), !. 

% for maximisation
down_class2(C,C,Test,_,Rels) :- return_solution(Test), !,  % If a solution, stop and maybe return C
   includable(C), rel_classes(Rels,Rel), member(C,Rel), !. 
down_class2(C,Hole,Test,KB,Rel) :-
   \+ \+ ((C=Hole, try_descendents(Test))), 
   all_children(KB,C,List),
   member(_C,List),
   \+ _C='BOTTOM',
   \+ \+((Hole=_C,on_path(Test))), 
   \+ \+ (down_class2(_C,Hole,Test,KB,Rel)), !,
   member(C1,List),
   \+ C1='BOTTOM',
   \+ \+((Hole=C1,on_path(Test))), 
   down_class2(C1,Hole,Test,KB,Rel).

% up_class(C,Hole,Test,KB,Rel)
% Optimised treatment of selection of an atomic class in a negative
% context

up_class(C,Hole,Test,KB,Rel) :-
   (most_general(Test) -> up_class2(C,Hole,Test,KB,Rel); up_class1(C,Hole,Test,KB,Rel)).

% for minimisation
up_class1(C,Hole,Test,KB,Rel) :-
   \+ \+ ((C=Hole, try_descendents(Test))),
   all_parents(KB,C,List),
   member(_C,List),
   \+ _C='TOP', 
   \+ \+((Hole=_C,on_path(Test))), 
   \+ \+ (up_class1(_C,Hole,Test,KB,Rel)), !,
   member(C1,List),
   \+ C1='TOP', 
   \+ \+((Hole=C1,on_path(Test))), 
   up_class1(C1,Hole,Test,KB,Rel).
up_class1(C,C,Test,_,Rels) :- includable(C), rel_classes(Rels,Rel), member(C,Rel), return_solution(Test), !. 

% for maximisation
up_class2(C,C,Test,_,Rels) :- return_solution(Test), !,
   includable(C), rel_classes(Rels,Rel), member(C,Rel), !. 
up_class2(C,Hole,Test,KB,Rel) :-
   \+ \+ ((C=Hole, try_descendents(Test))),
   all_parents(KB,C,List),
   member(_C,List),
   \+ _C='TOP', 
   \+ \+((Hole=_C,on_path(Test))), 
   \+ \+ (up_class2(_C,Hole,Test,KB,Rel)), !,
   member(C1,List),
   \+ C1='TOP', 
   \+ \+((Hole=C1,on_path(Test))), 
   up_class2(C1,Hole,Test,KB,Rel).

% Find a best number restriction for some property

find_ns(A,NA,P,KB,Res,Test) :-
   \+ most_general(Test), !, 
   gfind_ns(P,KB,GMin,GMax),
   find_min(Res,Test,GMin,Min),
   find_max(Res,Test,Min,GMax,Max),
   minmax(Min,Max,A,NA,GMin,GMax,Res).
find_ns(A,NA,_,_,Res,Test) :-
   %write(Res), write(' in '), write(Test), nl,
   %write(min(1000000)), nl,
   \+ \+ ((Res=[min_cardinality(1000000)],on_path(Test),return_solution(Test))),
   mg_min(0,Res,Test,Min), Min>1,
   Res=[min_cardinality(Min)], NA is A-1.
   %write('return '), write(min(Min)), nl.
find_ns(A,NA,_,_,Res,Test) :-
   %write(max(0)), nl,
   \+ \+ ((Res=[max_cardinality(0)],on_path(Test),return_solution(Test))),
   %write(max(1000000)), nl,
   \+ ((Res=[max_cardinality(1000000)],on_path(Test),return_solution(Test))),
   mg_max(0,Res,Test,Max),
   (Max=0 -> Res=[cardinality(0)]; Res=[max_cardinality(Max)]),
   NA is A-1.
   %write('return '), write(max(Max)), nl.

find_max(Res,Test,Min,GMax,Max) :-
   GMax<1000000, !,
   find_max1(Res,Test,Min,Max).
find_max(Res,Test,Min,1000000,Max) :-
   \+ \+ ((Res=[max_cardinality(1000000)],on_path(Test),return_solution(Test))), !,
   find_max1(Res,Test,Min,Max).
find_max(_,_,_,_,1000000).

minmax(X,X,A,A,G1,G2,[cardinality(X)]) :- (X>G1; X<G2), !.
minmax(X,Y,A,NA,G1,G2,[min_cardinality(X),max_cardinality(Y)]) :-
   \+ X=Y, X>G1, Y<G2,
   X>1,   % don't include => 1, because it will be covered by E.
   A>=1, NA is A-1.
minmax(X,Y,A,A,G1,_,[min_cardinality(X)]) :- \+ X=Y, X>G1, X>1.
minmax(X,Y,A,A,_,G2,[max_cardinality(Y)]) :- \+ X=Y, Y<G2.

find_min(Res,Test,N,N2) :-
   \+ \+ ((Res=[min_cardinality(N)],try_descendents(Test))), 
   N1 is N+1,
   \+ \+ ((Res=[min_cardinality(N1)],on_path(Test))), !,
   find_min(Res,Test,N1,N2).
find_min(Res,Test,N,N) :-
   \+ \+ ((Res=[min_cardinality(N)],return_solution(Test))). 

find_max1(Res,Test,Min,Min) :-
   \+ \+ ((Res=[max_cardinality(Min)],on_path(Test),return_solution(Test))), !.
find_max1(Res,Test,Min,Max) :-
   Min1 is Min+1,
   find_max1(Res,Test,Min1,Max).

% find global number restrictions for a property

gfind_ns(P,KB,Min,Max) :-
   gfind_min(P,KB,0,Min),
   (subsumes(KB,restriction(P,[max_cardinality(1000000)]),'TOP') ->
      gfind_max(Min,P,KB,Max); Max=1000000).

gfind_min(P,KB,N,N2) :-
   N1 is N+1,
   subsumes(KB,restriction(P,[min_cardinality(N1)]),'TOP'), !,
   gfind_min(P,KB,N1,N2).
gfind_min(_,_,N,N).

gfind_max(Min,P,KB,Min) :-
   subsumes(KB,restriction(P,[max_cardinality(Min)]),'TOP'), !.
gfind_max(Min,P,KB,Max) :-
   Min1 is Min+1,
   gfind_max(Min1,P,KB,Max).

% for most general answers

mg_min(N,Res,Test,N) :-
   \+ \+ ((Res=[min_cardinality(N)],on_path(Test),return_solution(Test))), !.
mg_min(N,Res,Test,Min) :- 
   N1 is N+1,
   mg_min(N1,Res,Test,Min).

mg_max(N,Res,Test,Max) :-
   N1 is N+1,
   \+ \+ ((Res=[max_cardinality(N1)],on_path(Test),return_solution(Test))), !,
   %write(max(N1)), write(' succeeds'), nl,
   mg_max(N1,Res,Test,Max).
mg_max(N,_,_,N).

% size of a general formula 

size(A,S) :- size(1,A,S).

size(_,A,0) :- atomic(A), !.
size(D,[X],S) :- !, size(D,X,S).
size(D,[X|List],S) :- !, size(D,X,S1), size(D,List,S2), S is S1+S2+1.
size(D,intersection_of(L),S) :- !, size(D,L,S).
size(D,union_of(L),S) :- !, size(D,L,S).
size(D,complement_of(X),S) :- !, size(D,X,S1), S is S1+1.
size(D,restriction(_,L),S) :- !, size(D,L,S).
size(D,all_values_from(X),S) :- !, D1 is D+1, size(D1,X,S1), S is S1+D.
size(D,some_values_from(X),S) :- !, D1 is D+1, size(D1,X,S1), S is S1+D.
size(D,X,D1) :- cardinality(X), !, D1 is D+1.
size(_,X,_) :- write('***CANT FIND SIZE OF: '), write(X), nl, fail.





